
class Scene(Entity):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        Entity(parent=self, position=Vec3(-1.0334, 0, 3.0372), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-9.71, 0.001, 14.34), scale=Vec3(1, 2, 1), model='cube', origin=Vec3(0, -0.5, 0), shader='lit_with_shadows_shader', color=Color(1.0, 0.5, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-4.11364, 0.0665623, -0.246485), scale=Vec3(0.487548, 0.487548, 0.487548), model='rotation_gizmo_model', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.5, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(0.3745, 0.002, -6.6139), model='cube', shader='lit_with_shadows_shader', color=Color(0.0, 0.5166666507720947, 1.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-0.2384, 0.065, -0.8437), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.5, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(1.1839, 0.003, 8.2525), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(0.0499, 0.067, 7.5857), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.5, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(4.3956, 0.004, -2.4644), model='cube', shader='lit_with_shadows_shader', color=Color(0.0, 0.5166666507720947, 1.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(2.1659, 0.067, -2.8756), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.5, 1.0), collider='box', )
        Entity(parent=self, model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-1.4502, 0.069, -6.4631), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.5, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-6.1612, 0.0358467, 0.566014), scale=Vec3(0.487548, 0.487548, 0.487548), model='rotation_gizmo_model', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-4.74014, 0.0358467, 0.250619), scale=Vec3(0.487548, 0.975097, 0.487548), model='rotation_gizmo_model', shader='lit_with_shadows_shader', color=Color(1.0, 0.5, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-7.5011, 0.065, 3.3701), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.5, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-5.34992, 0.0363343, -1.00102), scale=Vec3(0.487548, 0.487548, 0.487548), model='rotation_gizmo_model', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-5.9028, 0.0675374, -1.32611), scale=Vec3(0.487548, 0.487548, 0.487548), model='rotation_gizmo_model', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.5, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-4.30705, 0.0368218, -0.0443969), scale=Vec3(0.487548, 0.487548, 0.487548), model='rotation_gizmo_model', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-4.85993, 0.0680249, -0.369494), scale=Vec3(0.487548, 0.487548, 0.487548), model='rotation_gizmo_model', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.5, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-4.17766, 0.0373094, -1.99166), scale=Vec3(0.487548, 0.487548, 0.487548), model='rotation_gizmo_model', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-4.73054, 0.0685125, -2.31676), scale=Vec3(0.487548, 0.487548, 0.487548), model='rotation_gizmo_model', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.5, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-5.41242, 0.0377969, -2.4111), scale=Vec3(0.487548, 0.487548, 0.487548), model='rotation_gizmo_model', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-5.9653, 0.069, -2.7362), scale=Vec3(0.487548, 0.487548, 0.487548), model='rotation_gizmo_model', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.5, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(0.202197, -0.0600002, 17.3612), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(0.9271, -0.067, 12.5934), scale=Vec3(1, 2, 1), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.5, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(0.911404, -0.0040004, 15.1795), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.5, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-0.3235, -0.066, 10.0263), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(0.1282, -0.003, 2.8922), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.5, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(1.8154, -0.065, 11.9884), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(0.6814, -0.001, 11.3216), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.5, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(2.0808, -0.064, 7.9944), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(0.9468, 0, 7.3276), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.5, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(1.3989, -0.064, 0.6668), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.5, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(3.1222, 0.002, 2.838), scale=Vec3(1, 2, 1), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.5, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(4.0105, 0.004, 2.2329), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(2.8765, 0.068, 1.5661), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.5, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(0.839, -0.066, 7.2207), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(3.7537, -0.066, 6.5738), scale=Vec3(1, 2, 1), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.5, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(2.5031, -0.065, 4.0067), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(4.642, -0.064, 5.9688), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(3.508, 0, 5.302), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.5, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(4.9074, -0.063, 1.9748), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(3.7734, 0.001, 1.308), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.5, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-4.61, -0.28, 3.043), scale=Vec3(9.028, 0.171, 94.37), model='cube', shader='lit_with_shadows_shader', texture='sky_default', color=Color(0.6709499955177307, 0.7099999785423279, 0.24140000343322754, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-8.72107, 0.004, -1.17382), model='cube', shader='lit_with_shadows_shader', color=Color(0.0, 0.5166666507720947, 1.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-11.0359, -0.064, 9.28498), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-9.99447, 0.002, 4.12858), scale=Vec3(1, 2, 1), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.5, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-9.10617, 0.004, 3.52348), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-10.2402, 0.068, 2.85668), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.5, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-9.36297, -0.066, 7.86438), scale=Vec3(1, 2, 1), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.5, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-10.6136, -0.065, 5.29728), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-8.47467, -0.064, 7.25938), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-9.60867, 0, 6.59258), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.5, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-8.20927, -0.063, 3.26538), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-9.34327, 0.001, 2.59858), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.5, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-8.78803, 0.00399695, -13.4254), model='cube', shader='lit_with_shadows_shader', color=Color(0.0, 0.5166666507720947, 1.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-11.1029, -0.0640031, -2.96662), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-10.0614, 0.00199695, -8.12302), scale=Vec3(1, 2, 1), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.5, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-9.17313, 0.00399695, -8.72812), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-10.3072, 0.0679969, -9.39492), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.5, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-9.42993, -0.0660031, -4.38722), scale=Vec3(1, 2, 1), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.5, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-10.6806, -0.0650031, -6.95432), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-8.54163, -0.0640031, -4.99222), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-9.67563, -3.05171e-06, -5.65902), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.5, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-8.27623, -0.0630031, -8.98622), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.0, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(-9.41023, 0.000996948, -9.65302), model='cube', shader='lit_with_shadows_shader', color=Color(1.0, 0.0, 0.5, 1.0), collider='box', )
        Entity(parent=self, position=Vec3(13.63, -6.53, 0), model='cube', shader='lit_with_shadows_shader', collider='box', )
        Entity(parent=self, position=Vec3(-4.61, 0.916, -37), scale=Vec3(1, 1.44, 1), model='sphere', shader='matcap_shader', texture='shore', color=Color(1.0, 0.23333333432674408, 0.0, 1.0), collider='box', )
